@@ -27,7 +27,7 @@
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *	$Id: system.c,v 1.37 2005/06/01 19:22:06 roland Exp $
+ *	$Id: system.c,v 1.43 2007/01/16 15:10:08 ldv Exp $
  */
 
 #include "defs.h"
@@ -49,10 +49,16 @@
 #define MS_BIND		4096
 #define MS_MOVE		8192
 #define MS_REC		16384
-#define MS_VERBOSE	32768
+#define MS_SILENT	32768
 #define MS_POSIXACL	(1<<16)	/* VFS does not apply the umask */
+#define MS_UNBINDABLE	(1<<17)	/* change to unbindable */
+#define MS_PRIVATE	(1<<18)	/* change to private */
+#define MS_SLAVE	(1<<19)	/* change to slave */
+#define MS_SHARED	(1<<20)	/* change to shared */
 #define MS_ACTIVE	(1<<30)
 #define MS_NOUSER	(1<<31)
+#define MS_MGC_VAL	0xc0ed0000	/* Magic flag number */
+#define MS_MGC_MSK	0xffff0000	/* Magic flag mask */
 
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -79,6 +85,7 @@
 #include <linux/sysctl.h>
 
 static const struct xlat mount_flags[] = {
+	{ MS_MGC_VAL,	"MS_MGC_VAL"	},
 	{ MS_RDONLY,	"MS_RDONLY"	},
 	{ MS_NOSUID,	"MS_NOSUID"	},
 	{ MS_NODEV,	"MS_NODEV"	},
@@ -91,44 +98,74 @@ static const struct xlat mount_flags[] = {
 	{ MS_BIND,	"MS_BIND"	},
 	{ MS_MOVE,	"MS_MOVE"	},
 	{ MS_REC,	"MS_REC"	},
-	{ MS_VERBOSE,	"MS_VERBOSE"	},
+	{ MS_SILENT,	"MS_SILENT"	},
 	{ MS_POSIXACL,	"MS_POSIXACL"	},
+	{ MS_UNBINDABLE,"MS_UNBINDABLE"	},
+	{ MS_PRIVATE,	"MS_PRIVATE"	},
+	{ MS_SLAVE,	"MS_SLAVE"	},
+	{ MS_SHARED,	"MS_SHARED"	},
 	{ MS_ACTIVE,	"MS_ACTIVE"	},
 	{ MS_NOUSER,	"MS_NOUSER"	},
 	{ 0,		NULL		},
 };
 
 int
-sys_mount(tcp)
-struct tcb *tcp;
+sys_mount(struct tcb *tcp)
 {
 	if (entering(tcp)) {
+		int ignore_type = 0, ignore_data = 0;
+		unsigned long flags = tcp->u_arg[3];
+
+		/* Discard magic */
+		if ((flags & MS_MGC_MSK) == MS_MGC_VAL)
+			flags &= ~MS_MGC_MSK;
+
+		if (flags & MS_REMOUNT)
+			ignore_type = 1;
+		else if (flags & (MS_BIND | MS_MOVE))
+			ignore_type = ignore_data = 1;
+
 		printpath(tcp, tcp->u_arg[0]);
 		tprintf(", ");
+
 		printpath(tcp, tcp->u_arg[1]);
 		tprintf(", ");
-		if ((tcp->u_arg[3] & (MS_BIND|MS_MOVE|MS_REMOUNT)) == 0)
-			printpath(tcp, tcp->u_arg[2]);
-		else
+
+		if (ignore_type && tcp->u_arg[2])
 			tprintf("%#lx", tcp->u_arg[2]);
+		else
+			printstr(tcp, tcp->u_arg[2], -1);
 		tprintf(", ");
+
 		printflags(mount_flags, tcp->u_arg[3], "MS_???");
-		tprintf(", %#lx", tcp->u_arg[4]);
+		tprintf(", ");
+
+		if (ignore_data && tcp->u_arg[4])
+			tprintf("%#lx", tcp->u_arg[4]);
+		else
+			printstr(tcp, tcp->u_arg[4], -1);
 	}
 	return 0;
 }
 
+#define MNT_FORCE	0x00000001	/* Attempt to forcibily umount */
+#define MNT_DETACH	0x00000002	/* Just detach from the tree */
+#define MNT_EXPIRE	0x00000004	/* Mark for expiry */
+
+static const struct xlat umount_flags[] = {
+	{ MNT_FORCE,	"MNT_FORCE"	},
+	{ MNT_DETACH,	"MNT_DETACH"	},
+	{ MNT_EXPIRE,	"MNT_EXPIRE"	},
+	{ 0,		NULL		},
+};
+
 int
-sys_umount2(tcp)
-struct tcb *tcp;
+sys_umount2(struct tcb *tcp)
 {
 	if (entering(tcp)) {
 		printstr(tcp, tcp->u_arg[0], -1);
 		tprintf(", ");
-		if (tcp->u_arg[1] & 1)
-			tprintf("MNT_FORCE");
-		else
-			tprintf("0");
+		printflags(umount_flags, tcp->u_arg[1], "MNT_???");
 	}
 	return 0;
 }
@@ -1568,6 +1605,9 @@ struct tcb *tcp;
 #endif
 
 #ifdef LINUX
+/* Linux 2.6.18+ headers removed CTL_PROC enum.  */
+# define CTL_PROC 4
+# define CTL_CPU 10		/* older headers lack */
 static const struct xlat sysctl_root[] = {
 	{ CTL_KERN, "CTL_KERN" },
 	{ CTL_VM, "CTL_VM" },
@@ -1576,6 +1616,9 @@ static const struct xlat sysctl_root[] = {
 	{ CTL_FS, "CTL_FS" },
 	{ CTL_DEBUG, "CTL_DEBUG" },
 	{ CTL_DEV, "CTL_DEV" },
+	{ CTL_BUS, "CTL_BUS" },
+	{ CTL_ABI, "CTL_ABI" },
+	{ CTL_CPU, "CTL_CPU" },
 	{ 0, NULL }
 };
 
@@ -1921,7 +1964,9 @@ struct tcb *tcp;
 			goto out;
 		}
 	out:
-		max_cnt = abbrev(tcp) ? max_strlen : info.nlen;
+		max_cnt = info.nlen;
+		if (abbrev(tcp) && max_cnt > max_strlen)
+			max_cnt = max_strlen;
 		while (cnt < max_cnt)
 			tprintf(", %x", name[cnt++]);
 		if (cnt < info.nlen)
